from PyQt6.QtCore import QTimer, pyqtSlot, Qt, pyqtSignal
from PyQt6.QtGui import (
    QIcon,
    QKeySequence,
    QWheelEvent,
    QCursor,
    QContextMenuEvent,
    QAction,
)
from PyQt6.QtWidgets import QMenu

from urh.ui.painting.SceneManager import SceneManager
from urh.ui.views.SelectableGraphicView import SelectableGraphicView
from urh.util.Logger import logger


class ZoomableGraphicView(SelectableGraphicView):
    MINIMUM_VIEW_WIDTH = 300

    # argument is x zoom factor
    # if argument is -1, then show_full_scene was triggered during zoom
    zoomed = pyqtSignal(float)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.context_menu_position = None  # type: QPoint
        self.scene_type = 0

        self.auto_fit_on_resize_is_blocked = False

        self.zoom_in_action = QAction(self.tr("Zoom in"), self)
        self.zoom_in_action.setShortcut(QKeySequence.StandardKey.ZoomIn)
        self.zoom_in_action.triggered.connect(self.on_zoom_in_action_triggered)
        self.zoom_in_action.setShortcutContext(
            Qt.ShortcutContext.WidgetWithChildrenShortcut
        )
        self.zoom_in_action.setIcon(QIcon.fromTheme("zoom-in"))
        self.addAction(self.zoom_in_action)

        self.zoom_out_action = QAction(self.tr("Zoom out"), self)
        self.zoom_out_action.setShortcut(QKeySequence.StandardKey.ZoomOut)
        self.zoom_out_action.triggered.connect(self.on_zoom_out_action_triggered)
        self.zoom_out_action.setShortcutContext(
            Qt.ShortcutContext.WidgetWithChildrenShortcut
        )
        self.zoom_out_action.setIcon(QIcon.fromTheme("zoom-out"))
        self.addAction(self.zoom_out_action)

        self.zoom_original_action = QAction(self.tr("Zoom original"), self)
        self.zoom_original_action.setShortcut("Ctrl+0")
        self.zoom_original_action.triggered.connect(
            self.on_zoom_original_action_triggered
        )
        self.zoom_original_action.setShortcutContext(
            Qt.ShortcutContext.WidgetWithChildrenShortcut
        )
        self.zoom_original_action.setIcon(QIcon.fromTheme("zoom-original"))
        self.addAction(self.zoom_original_action)

        self.redraw_timer = QTimer(self)
        self.redraw_timer.setSingleShot(True)
        self.redraw_timer.timeout.connect(self.redraw_view)

        self.zoomed.connect(self.on_signal_zoomed)

        self.scene_x_zoom_stretch = 1

    @property
    def y_center(self):
        try:
            if self.scene_type == 0:
                # Normal scene
                return 0
            else:
                return -self.signal.center
        except Exception as e:
            logger.error(
                "Could not access y_center property: {0}. Falling back to 0".format(e)
            )
            return 0

    def create_context_menu(self):
        menu = QMenu()
        self._add_zoom_actions_to_menu(menu)
        return menu

    def contextMenuEvent(self, event: QContextMenuEvent):
        self.context_menu_position = event.pos()
        menu = self.create_context_menu()
        menu.exec(self.mapToGlobal(event.pos()))
        self.context_menu_position = None

    def _add_zoom_actions_to_menu(self, menu: QMenu):
        menu.addAction(self.zoom_in_action)
        menu.addAction(self.zoom_out_action)

        if self.something_is_selected:
            zoom_action = menu.addAction(self.tr("Zoom selection"))
            zoom_action.setIcon(QIcon.fromTheme("zoom-fit-best"))
            zoom_action.triggered.connect(self.on_zoom_action_triggered)

        menu.addSeparator()

    def scrollContentsBy(self, dx: int, dy: int):
        try:
            super().scrollContentsBy(dx, dy)
            self.redraw_timer.start(0)
        except RuntimeError as e:
            logger.warning("Graphic View already closed: " + str(e))

    def zoom(self, factor, zoom_to_mouse_cursor=True, cursor_pos=None):
        if factor > 1 and self.view_rect().width() / factor < self.MINIMUM_VIEW_WIDTH:
            factor = self.view_rect().width() / self.MINIMUM_VIEW_WIDTH

        if zoom_to_mouse_cursor:
            pos = (
                self.mapFromGlobal(QCursor.pos()) if cursor_pos is None else cursor_pos
            )
        else:
            pos = None
        old_pos = self.mapToScene(pos) if pos is not None else None

        show_full = False
        if self.view_rect().width() / factor > self.sceneRect().width():
            self.show_full_scene()
            factor = 1
            show_full = True

        self.scale(factor, 1)
        if show_full:
            self.zoomed.emit(-1)
        else:
            self.zoomed.emit(factor)

        if pos is not None:
            move = self.mapToScene(pos) - old_pos
            self.translate(move.x(), 0)

    def wheelEvent(self, event: QWheelEvent):
        zoom_factor = 1.001 ** event.angleDelta().y()
        self.zoom(zoom_factor, cursor_pos=event.position().toPoint())

    def resizeEvent(self, event):
        if self.sceneRect().width() == 0 or self.auto_fit_on_resize_is_blocked:
            return

        self.auto_fit_view()

    def auto_fit_view(self):
        h_tar = self.sceneRect().height()
        h_view = self.view_rect().height()

        if abs(h_tar) > 0:
            self.scale(1, h_view / h_tar)
        self.centerOn(
            self.view_rect().x() + self.view_rect().width() / 2, self.y_center
        )

    def show_full_scene(self, reinitialize=False):
        y_factor = self.transform().m22()
        self.resetTransform()
        # Use full self.width() here to enable show_full_scene when view_rect not yet set e.g. in Record Signal Dialog
        x_factor = (
            self.width() / (self.sceneRect().width() * self.scene_x_zoom_stretch)
            if self.sceneRect().width()
            else 1
        )
        self.scale(x_factor, y_factor)
        self.centerOn(0, self.y_center)

        self.redraw_view(reinitialize)

    def zoom_to_selection(self, start: int, end: int):
        if start == end:
            return

        x_factor = self.view_rect().width() / (end - start)
        self.zoom(x_factor, zoom_to_mouse_cursor=False)
        self.centerOn(start + (end - start) / 2, self.y_center)

    def plot_data(self, data):
        if self.scene_manager is None:
            self.scene_manager = SceneManager(self)

        self.scene_manager.plot_data = data
        self.scene_manager.init_scene()
        self.setScene(self.scene_manager.scene)
        self.scene_manager.show_full_scene()

    def redraw_view(self, reinitialize=False):
        if hasattr(self, "scene_manager") and self.scene_manager is not None:
            self.scene_manager.scene_type = self.scene_type
            if reinitialize:
                self.scene_manager.init_scene()

            vr = self.view_rect()
            start, end = vr.x(), vr.x() + vr.width()
            self.scene_manager.show_scene_section(
                start, end, *self._get_sub_path_ranges_and_colors(start, end)
            )

            if self.scene_type == 1:
                self.scene().redraw_legend()
            else:
                self.scene().hide_legend()

    def _get_sub_path_ranges_and_colors(self, start: float, end: float):
        # Overwritten in Epic Graphic View
        return None, None

    def eliminate(self):
        self.redraw_timer.stop()
        super().eliminate()

    @pyqtSlot()
    def on_signal_zoomed(self):
        self.redraw_timer.start(30)

    @pyqtSlot()
    def on_zoom_in_action_triggered(self):
        self.zoom(1.1)

    @pyqtSlot()
    def on_zoom_out_action_triggered(self):
        self.zoom(0.9)

    @pyqtSlot()
    def on_zoom_original_action_triggered(self):
        self.show_full_scene(reinitialize=False)
        self.zoomed.emit(-1)

    @pyqtSlot()
    def on_zoom_action_triggered(self):
        self.zoom_to_selection(self.selection_area.start, self.selection_area.end)



#
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_FuzzingDialog(object):
    def setupUi(self, FuzzingDialog):
        FuzzingDialog.setObjectName("FuzzingDialog")
        FuzzingDialog.resize(523, 632)
        self.gridLayout_5 = QtWidgets.QGridLayout(FuzzingDialog)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.spinBoxFuzzMessage = QtWidgets.QSpinBox(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.spinBoxFuzzMessage.sizePolicy().hasHeightForWidth()
        )
        self.spinBoxFuzzMessage.setSizePolicy(sizePolicy)
        self.spinBoxFuzzMessage.setMaximum(999999999)
        self.spinBoxFuzzMessage.setObjectName("spinBoxFuzzMessage")
        self.gridLayout_5.addWidget(self.spinBoxFuzzMessage, 2, 1, 1, 1)
        self.comboBoxStrategy = QtWidgets.QComboBox(parent=FuzzingDialog)
        self.comboBoxStrategy.setObjectName("comboBoxStrategy")
        self.comboBoxStrategy.addItem("")
        self.comboBoxStrategy.addItem("")
        self.comboBoxStrategy.addItem("")
        self.comboBoxStrategy.addItem("")
        self.gridLayout_5.addWidget(self.comboBoxStrategy, 6, 1, 1, 1)
        self.btnAddFuzzingValues = QtWidgets.QPushButton(parent=FuzzingDialog)
        self.btnAddFuzzingValues.setObjectName("btnAddFuzzingValues")
        self.gridLayout_5.addWidget(self.btnAddFuzzingValues, 9, 1, 1, 1)
        self.stackedWidgetLabels = QtWidgets.QStackedWidget(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.stackedWidgetLabels.sizePolicy().hasHeightForWidth()
        )
        self.stackedWidgetLabels.setSizePolicy(sizePolicy)
        self.stackedWidgetLabels.setObjectName("stackedWidgetLabels")
        self.pageAddRangeLabel = QtWidgets.QWidget()
        self.pageAddRangeLabel.setObjectName("pageAddRangeLabel")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.pageAddRangeLabel)
        self.verticalLayout_3.setContentsMargins(0, 0, -1, 0)
        self.verticalLayout_3.setSpacing(6)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.lStart = QtWidgets.QLabel(parent=self.pageAddRangeLabel)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lStart.sizePolicy().hasHeightForWidth())
        self.lStart.setSizePolicy(sizePolicy)
        self.lStart.setObjectName("lStart")
        self.verticalLayout_3.addWidget(self.lStart)
        self.lEnd = QtWidgets.QLabel(parent=self.pageAddRangeLabel)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lEnd.sizePolicy().hasHeightForWidth())
        self.lEnd.setSizePolicy(sizePolicy)
        self.lEnd.setObjectName("lEnd")
        self.verticalLayout_3.addWidget(self.lEnd)
        self.lStep = QtWidgets.QLabel(parent=self.pageAddRangeLabel)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lStep.sizePolicy().hasHeightForWidth())
        self.lStep.setSizePolicy(sizePolicy)
        self.lStep.setObjectName("lStep")
        self.verticalLayout_3.addWidget(self.lStep)
        self.stackedWidgetLabels.addWidget(self.pageAddRangeLabel)
        self.pageAddBoundariesLabel = QtWidgets.QWidget()
        self.pageAddBoundariesLabel.setObjectName("pageAddBoundariesLabel")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.pageAddBoundariesLabel)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setSpacing(6)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.checkBoxLowerBound = QtWidgets.QCheckBox(
            parent=self.pageAddBoundariesLabel
        )
        self.checkBoxLowerBound.setChecked(True)
        self.checkBoxLowerBound.setObjectName("checkBoxLowerBound")
        self.verticalLayout_4.addWidget(self.checkBoxLowerBound)
        self.checkBoxUpperBound = QtWidgets.QCheckBox(
            parent=self.pageAddBoundariesLabel
        )
        self.checkBoxUpperBound.setChecked(True)
        self.checkBoxUpperBound.setObjectName("checkBoxUpperBound")
        self.verticalLayout_4.addWidget(self.checkBoxUpperBound)
        self.lNumberBoundaries = QtWidgets.QLabel(parent=self.pageAddBoundariesLabel)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.lNumberBoundaries.sizePolicy().hasHeightForWidth()
        )
        self.lNumberBoundaries.setSizePolicy(sizePolicy)
        self.lNumberBoundaries.setObjectName("lNumberBoundaries")
        self.verticalLayout_4.addWidget(self.lNumberBoundaries)
        self.stackedWidgetLabels.addWidget(self.pageAddBoundariesLabel)
        self.pageAddRandomLabel = QtWidgets.QWidget()
        self.pageAddRandomLabel.setObjectName("pageAddRandomLabel")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.pageAddRandomLabel)
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_6.setSpacing(6)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.lRandomMin = QtWidgets.QLabel(parent=self.pageAddRandomLabel)
        self.lRandomMin.setObjectName("lRandomMin")
        self.verticalLayout_6.addWidget(self.lRandomMin)
        self.lRandomMax = QtWidgets.QLabel(parent=self.pageAddRandomLabel)
        self.lRandomMax.setObjectName("lRandomMax")
        self.verticalLayout_6.addWidget(self.lRandomMax)
        self.lNumRandom = QtWidgets.QLabel(parent=self.pageAddRandomLabel)
        self.lNumRandom.setObjectName("lNumRandom")
        self.verticalLayout_6.addWidget(self.lNumRandom)
        self.stackedWidgetLabels.addWidget(self.pageAddRandomLabel)
        self.gridLayout_5.addWidget(self.stackedWidgetLabels, 7, 0, 2, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.lPreBits = QtWidgets.QLabel(parent=FuzzingDialog)
        self.lPreBits.setObjectName("lPreBits")
        self.horizontalLayout_2.addWidget(self.lPreBits)
        self.lFuzzedBits = QtWidgets.QLabel(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lFuzzedBits.sizePolicy().hasHeightForWidth())
        self.lFuzzedBits.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lFuzzedBits.setFont(font)
        self.lFuzzedBits.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.lFuzzedBits.setObjectName("lFuzzedBits")
        self.horizontalLayout_2.addWidget(self.lFuzzedBits)
        self.lPostBits = QtWidgets.QLabel(parent=FuzzingDialog)
        self.lPostBits.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignRight
            | QtCore.Qt.AlignmentFlag.AlignTrailing
            | QtCore.Qt.AlignmentFlag.AlignVCenter
        )
        self.lPostBits.setObjectName("lPostBits")
        self.horizontalLayout_2.addWidget(self.lPostBits)
        spacerItem = QtWidgets.QSpacerItem(
            40,
            20,
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Minimum,
        )
        self.horizontalLayout_2.addItem(spacerItem)
        self.gridLayout_5.addLayout(self.horizontalLayout_2, 1, 1, 1, 1)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.lFuzzedValues = QtWidgets.QLabel(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.lFuzzedValues.sizePolicy().hasHeightForWidth()
        )
        self.lFuzzedValues.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lFuzzedValues.setFont(font)
        self.lFuzzedValues.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.lFuzzedValues.setObjectName("lFuzzedValues")
        self.verticalLayout.addWidget(self.lFuzzedValues)
        self.chkBRemoveDuplicates = QtWidgets.QCheckBox(parent=FuzzingDialog)
        self.chkBRemoveDuplicates.setObjectName("chkBRemoveDuplicates")
        self.verticalLayout.addWidget(self.chkBRemoveDuplicates)
        self.gridLayout_4 = QtWidgets.QGridLayout()
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.btnDelRow = QtWidgets.QToolButton(parent=FuzzingDialog)
        icon = QtGui.QIcon.fromTheme("list-remove")
        self.btnDelRow.setIcon(icon)
        self.btnDelRow.setObjectName("btnDelRow")
        self.gridLayout_4.addWidget(self.btnDelRow, 1, 1, 1, 1)
        self.tblFuzzingValues = FuzzingTableView(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.tblFuzzingValues.sizePolicy().hasHeightForWidth()
        )
        self.tblFuzzingValues.setSizePolicy(sizePolicy)
        self.tblFuzzingValues.setAlternatingRowColors(True)
        self.tblFuzzingValues.setVerticalScrollMode(
            QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel
        )
        self.tblFuzzingValues.setHorizontalScrollMode(
            QtWidgets.QAbstractItemView.ScrollMode.ScrollPerPixel
        )
        self.tblFuzzingValues.setShowGrid(False)
        self.tblFuzzingValues.setObjectName("tblFuzzingValues")
        self.tblFuzzingValues.horizontalHeader().setHighlightSections(False)
        self.tblFuzzingValues.verticalHeader().setHighlightSections(False)
        self.gridLayout_4.addWidget(self.tblFuzzingValues, 0, 0, 4, 1)
        spacerItem1 = QtWidgets.QSpacerItem(
            20,
            40,
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        self.gridLayout_4.addItem(spacerItem1, 3, 1, 1, 1)
        self.btnAddRow = QtWidgets.QToolButton(parent=FuzzingDialog)
        icon = QtGui.QIcon.fromTheme("list-add")
        self.btnAddRow.setIcon(icon)
        self.btnAddRow.setObjectName("btnAddRow")
        self.gridLayout_4.addWidget(self.btnAddRow, 0, 1, 1, 1)
        self.btnRepeatValues = QtWidgets.QToolButton(parent=FuzzingDialog)
        icon = QtGui.QIcon.fromTheme("media-playlist-repeat")
        self.btnRepeatValues.setIcon(icon)
        self.btnRepeatValues.setObjectName("btnRepeatValues")
        self.gridLayout_4.addWidget(self.btnRepeatValues, 2, 1, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout_4)
        self.gridLayout_5.addLayout(self.verticalLayout, 5, 0, 1, 2)
        self.stackedWidgetSpinboxes = QtWidgets.QStackedWidget(parent=FuzzingDialog)
        self.stackedWidgetSpinboxes.setObjectName("stackedWidgetSpinboxes")
        self.pageAddRange = QtWidgets.QWidget()
        self.pageAddRange.setObjectName("pageAddRange")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.pageAddRange)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.sBAddRangeStart = QtWidgets.QSpinBox(parent=self.pageAddRange)
        self.sBAddRangeStart.setObjectName("sBAddRangeStart")
        self.verticalLayout_2.addWidget(self.sBAddRangeStart)
        self.sBAddRangeEnd = QtWidgets.QSpinBox(parent=self.pageAddRange)
        self.sBAddRangeEnd.setObjectName("sBAddRangeEnd")
        self.verticalLayout_2.addWidget(self.sBAddRangeEnd)
        self.sBAddRangeStep = QtWidgets.QSpinBox(parent=self.pageAddRange)
        self.sBAddRangeStep.setMinimum(1)
        self.sBAddRangeStep.setObjectName("sBAddRangeStep")
        self.verticalLayout_2.addWidget(self.sBAddRangeStep)
        self.stackedWidgetSpinboxes.addWidget(self.pageAddRange)
        self.pageAddBoundaries = QtWidgets.QWidget()
        self.pageAddBoundaries.setObjectName("pageAddBoundaries")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.pageAddBoundaries)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setSpacing(6)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.spinBoxLowerBound = QtWidgets.QSpinBox(parent=self.pageAddBoundaries)
        self.spinBoxLowerBound.setObjectName("spinBoxLowerBound")
        self.verticalLayout_5.addWidget(self.spinBoxLowerBound)
        self.spinBoxUpperBound = QtWidgets.QSpinBox(parent=self.pageAddBoundaries)
        self.spinBoxUpperBound.setProperty("value", 13)
        self.spinBoxUpperBound.setObjectName("spinBoxUpperBound")
        self.verticalLayout_5.addWidget(self.spinBoxUpperBound)
        self.spinBoxBoundaryNumber = QtWidgets.QSpinBox(parent=self.pageAddBoundaries)
        self.spinBoxBoundaryNumber.setMinimum(1)
        self.spinBoxBoundaryNumber.setObjectName("spinBoxBoundaryNumber")
        self.verticalLayout_5.addWidget(self.spinBoxBoundaryNumber)
        self.stackedWidgetSpinboxes.addWidget(self.pageAddBoundaries)
        self.pageAddRandom = QtWidgets.QWidget()
        self.pageAddRandom.setObjectName("pageAddRandom")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.pageAddRandom)
        self.verticalLayout_7.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_7.setSpacing(6)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.spinBoxRandomMinimum = QtWidgets.QSpinBox(parent=self.pageAddRandom)
        self.spinBoxRandomMinimum.setObjectName("spinBoxRandomMinimum")
        self.verticalLayout_7.addWidget(self.spinBoxRandomMinimum)
        self.spinBoxRandomMaximum = QtWidgets.QSpinBox(parent=self.pageAddRandom)
        self.spinBoxRandomMaximum.setProperty("value", 42)
        self.spinBoxRandomMaximum.setObjectName("spinBoxRandomMaximum")
        self.verticalLayout_7.addWidget(self.spinBoxRandomMaximum)
        self.spinBoxNumberRandom = QtWidgets.QSpinBox(parent=self.pageAddRandom)
        self.spinBoxNumberRandom.setMinimum(1)
        self.spinBoxNumberRandom.setMaximum(999999999)
        self.spinBoxNumberRandom.setObjectName("spinBoxNumberRandom")
        self.verticalLayout_7.addWidget(self.spinBoxNumberRandom)
        self.stackedWidgetSpinboxes.addWidget(self.pageAddRandom)
        self.gridLayout_5.addWidget(self.stackedWidgetSpinboxes, 7, 1, 2, 1)
        self.lSourceBlock = QtWidgets.QLabel(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Maximum
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lSourceBlock.sizePolicy().hasHeightForWidth())
        self.lSourceBlock.setSizePolicy(sizePolicy)
        self.lSourceBlock.setObjectName("lSourceBlock")
        self.gridLayout_5.addWidget(self.lSourceBlock, 1, 0, 1, 1)
        self.lFuzzingReferenceBlock = QtWidgets.QLabel(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.lFuzzingReferenceBlock.sizePolicy().hasHeightForWidth()
        )
        self.lFuzzingReferenceBlock.setSizePolicy(sizePolicy)
        self.lFuzzingReferenceBlock.setObjectName("lFuzzingReferenceBlock")
        self.gridLayout_5.addWidget(self.lFuzzingReferenceBlock, 2, 0, 1, 1)
        self.lFuzzingEnd = QtWidgets.QLabel(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Maximum
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lFuzzingEnd.sizePolicy().hasHeightForWidth())
        self.lFuzzingEnd.setSizePolicy(sizePolicy)
        self.lFuzzingEnd.setObjectName("lFuzzingEnd")
        self.gridLayout_5.addWidget(self.lFuzzingEnd, 4, 0, 1, 1)
        self.lStrategy = QtWidgets.QLabel(parent=FuzzingDialog)
        self.lStrategy.setObjectName("lStrategy")
        self.gridLayout_5.addWidget(self.lStrategy, 6, 0, 1, 1)
        self.comboBoxFuzzingLabel = QtWidgets.QComboBox(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.comboBoxFuzzingLabel.sizePolicy().hasHeightForWidth()
        )
        self.comboBoxFuzzingLabel.setSizePolicy(sizePolicy)
        self.comboBoxFuzzingLabel.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.comboBoxFuzzingLabel.setEditable(True)
        self.comboBoxFuzzingLabel.setObjectName("comboBoxFuzzingLabel")
        self.gridLayout_5.addWidget(self.comboBoxFuzzingLabel, 0, 1, 1, 1)
        self.spinBoxFuzzingEnd = QtWidgets.QSpinBox(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.spinBoxFuzzingEnd.sizePolicy().hasHeightForWidth()
        )
        self.spinBoxFuzzingEnd.setSizePolicy(sizePolicy)
        self.spinBoxFuzzingEnd.setMinimum(1)
        self.spinBoxFuzzingEnd.setMaximum(999999999)
        self.spinBoxFuzzingEnd.setObjectName("spinBoxFuzzingEnd")
        self.gridLayout_5.addWidget(self.spinBoxFuzzingEnd, 4, 1, 1, 1)
        self.spinBoxFuzzingStart = QtWidgets.QSpinBox(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.spinBoxFuzzingStart.sizePolicy().hasHeightForWidth()
        )
        self.spinBoxFuzzingStart.setSizePolicy(sizePolicy)
        self.spinBoxFuzzingStart.setMinimum(1)
        self.spinBoxFuzzingStart.setMaximum(999999999)
        self.spinBoxFuzzingStart.setObjectName("spinBoxFuzzingStart")
        self.gridLayout_5.addWidget(self.spinBoxFuzzingStart, 3, 1, 1, 1)
        self.lFuzzingStart = QtWidgets.QLabel(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Maximum
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.lFuzzingStart.sizePolicy().hasHeightForWidth()
        )
        self.lFuzzingStart.setSizePolicy(sizePolicy)
        self.lFuzzingStart.setObjectName("lFuzzingStart")
        self.gridLayout_5.addWidget(self.lFuzzingStart, 3, 0, 1, 1)
        self.lFuzzingLabel = QtWidgets.QLabel(parent=FuzzingDialog)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.lFuzzingLabel.sizePolicy().hasHeightForWidth()
        )
        self.lFuzzingLabel.setSizePolicy(sizePolicy)
        self.lFuzzingLabel.setObjectName("lFuzzingLabel")
        self.gridLayout_5.addWidget(self.lFuzzingLabel, 0, 0, 1, 1)

        self.retranslateUi(FuzzingDialog)
        self.stackedWidgetLabels.setCurrentIndex(0)
        self.stackedWidgetSpinboxes.setCurrentIndex(0)
        self.comboBoxStrategy.currentIndexChanged["int"].connect(self.stackedWidgetLabels.setCurrentIndex)  # type: ignore
        self.comboBoxStrategy.currentIndexChanged["int"].connect(self.stackedWidgetSpinboxes.setCurrentIndex)  # type: ignore

    def retranslateUi(self, FuzzingDialog):
        _translate = QtCore.QCoreApplication.translate
        FuzzingDialog.setWindowTitle(_translate("FuzzingDialog", "Fuzzing"))
        self.comboBoxStrategy.setItemText(
            0, _translate("FuzzingDialog", "Add Range of Values")
        )
        self.comboBoxStrategy.setItemText(
            1, _translate("FuzzingDialog", "Add Boundaries")
        )
        self.comboBoxStrategy.setItemText(
            2, _translate("FuzzingDialog", "Add Random Values from Range")
        )
        self.comboBoxStrategy.setItemText(
            3, _translate("FuzzingDialog", "Add De Bruijn Sequence")
        )
        self.btnAddFuzzingValues.setText(
            _translate("FuzzingDialog", "Add to Fuzzed Values")
        )
        self.lStart.setText(_translate("FuzzingDialog", "Start (Decimal):"))
        self.lEnd.setText(_translate("FuzzingDialog", "End (Decimal):"))
        self.lStep.setText(_translate("FuzzingDialog", "Step (Decimal):"))
        self.checkBoxLowerBound.setText(_translate("FuzzingDialog", "Lower Bound"))
        self.checkBoxUpperBound.setText(_translate("FuzzingDialog", "Upper Bound"))
        self.lNumberBoundaries.setText(
            _translate("FuzzingDialog", "Values per Boundary:")
        )
        self.lRandomMin.setText(_translate("FuzzingDialog", "Range Minimum:"))
        self.lRandomMax.setText(_translate("FuzzingDialog", "Range Maximum:"))
        self.lNumRandom.setText(_translate("FuzzingDialog", "Number Values:"))
        self.lPreBits.setText(_translate("FuzzingDialog", "1111"))
        self.lFuzzedBits.setText(_translate("FuzzingDialog", "1010"))
        self.lPostBits.setText(_translate("FuzzingDialog", "010101"))
        self.lFuzzedValues.setText(_translate("FuzzingDialog", "Fuzzed Values"))
        self.chkBRemoveDuplicates.setText(
            _translate("FuzzingDialog", "Remove Duplicates")
        )
        self.btnDelRow.setToolTip(
            _translate(
                "FuzzingDialog",
                "Remove selected values or last value if nothing is selected.",
            )
        )
        self.btnDelRow.setText(_translate("FuzzingDialog", "..."))
        self.btnAddRow.setToolTip(_translate("FuzzingDialog", "Add a new value."))
        self.btnAddRow.setText(_translate("FuzzingDialog", "..."))
        self.btnRepeatValues.setToolTip(
            _translate(
                "FuzzingDialog",
                "Repeat selected values or all values if nothing is selected.",
            )
        )
        self.btnRepeatValues.setText(_translate("FuzzingDialog", "..."))
        self.lSourceBlock.setText(_translate("FuzzingDialog", "Source Message:"))
        self.lFuzzingReferenceBlock.setText(
            _translate("FuzzingDialog", "Message to fuzz:")
        )
        self.lFuzzingEnd.setText(
            _translate("FuzzingDialog", "Fuzzing Label End Index:")
        )
        self.lStrategy.setText(_translate("FuzzingDialog", "Strategy:"))
        self.lFuzzingStart.setText(
            _translate("FuzzingDialog", "Fuzzing Label Start Index:")
        )
        self.lFuzzingLabel.setText(_translate("FuzzingDialog", "Fuzzing Label:"))


from urh.ui.views.FuzzingTableView import FuzzingTableView



from PyQt6.QtCore import QModelIndex, Qt
from PyQt6.QtGui import QIcon

from urh.models.ProtocolTreeModel import ProtocolTreeModel


class GeneratorTreeModel(ProtocolTreeModel):
    def __init__(self, controller, parent=None):
        super().__init__(controller, parent)

    def set_root_item(self, root_item):
        self.rootItem = root_item

    def flags(self, index: QModelIndex):
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled

        return (
            Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsDragEnabled
            | Qt.ItemFlag.ItemIsDropEnabled
        )

    def mimeTypes(self):
        return []  # Prohibit Drag Drop in Generator

    def data(self, index: QModelIndex, role=None):
        item = self.getItem(index)
        if role == Qt.ItemDataRole.DisplayRole:  #
            return item.data()
        elif role == Qt.ItemDataRole.DecorationRole and item.is_group:
            return QIcon.fromTheme("folder")



from array import array

import numpy
from PyQt6.QtCore import Qt, pyqtSlot, QRegularExpression, QTimer
from PyQt6.QtGui import (
    QCloseEvent,
    QResizeEvent,
    QKeyEvent,
    QIcon,
    QRegularExpressionValidator,
)
from PyQt6.QtWidgets import QDialog, QMessageBox, QLineEdit

from urh import settings
from urh.controller.dialogs.ModulationParametersDialog import ModulationParametersDialog
from urh.signalprocessing.Modulator import Modulator
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.ui.ui_modulation import Ui_DialogModulation
from urh.util.Logger import logger


class ModulatorDialog(QDialog):
    def __init__(self, modulators, tree_model=None, parent=None):
        """
        :type modulators: list of Modulator
        """
        super().__init__(parent)

        self.ui = Ui_DialogModulation()
        self.ui.setupUi(self)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setWindowFlags(Qt.WindowType.Window)

        self.lock_samples_in_view = False

        if tree_model is not None:
            self.ui.treeViewSignals.setModel(tree_model)
            self.ui.treeViewSignals.expandAll()
            self.ui.gVOriginalSignal.signal_tree_root = tree_model.rootItem

        self.ui.comboBoxCustomModulations.clear()
        for modulator in modulators:
            self.ui.comboBoxCustomModulations.addItem(modulator.name)
        if len(modulators) == 1:
            self.ui.btnRemoveModulation.setDisabled(True)

        self.modulators = modulators

        self.set_ui_for_current_modulator()

        self.ui.cbShowDataBitsOnly.setText(self.tr("Show Only Data Sequence\n"))
        self.ui.cbShowDataBitsOnly.setEnabled(False)
        self.protocol = None  # type: ProtocolAnalyzer
        self.search_results = []
        self.ui.cbShowDataBitsOnly.setEnabled(False)
        self.ui.btnSearchNext.setEnabled(False)
        self.ui.btnSearchPrev.setEnabled(False)

        self.ui.chkBoxLockSIV.setDisabled(True)

        self.original_bits = ""

        self.restore_bits_action = self.ui.linEdDataBits.addAction(
            QIcon.fromTheme("edit-undo"), QLineEdit.ActionPosition.TrailingPosition
        )
        self.restore_bits_action.setEnabled(False)

        self.configure_parameters_action = self.ui.lineEditParameters.addAction(
            QIcon.fromTheme("configure"), QLineEdit.ActionPosition.TrailingPosition
        )

        self.create_connects()
        self.restoreGeometry(
            settings.read("{}/geometry".format(self.__class__.__name__), type=bytes)
        )

        self.set_bits_per_symbol_enabled_status()
        self.set_modulation_profile_status()

        # Ensure full screen is shown after resize
        QTimer.singleShot(100, self.show_full_scene)

    def __cur_selected_mod_type(self):
        s = self.ui.comboBoxModulationType.currentText()
        return s[s.rindex("(") + 1 : s.rindex(")")]

    @staticmethod
    def __trim_number(number):
        if abs(number) >= 1e9:  # giga
            return numpy.round(number / 1e9) * 1e9
        elif abs(number) >= 1e6:  # mega
            return numpy.round(number / 1e6) * 1e6
        elif abs(number) >= 1e3:  # Kilo
            return numpy.round(number / 1e3) * 1e3
        else:
            return number

    @staticmethod
    def __ensure_multitude(num1, num2):
        try:
            if abs(num1) > abs(num2):
                num1 = abs(int(num1 / num2)) * num2
            else:
                num2 = abs(int(num2 / num1)) * num1
            return num1, num2
        except Exception:
            return num1, num2

    def __set_gauss_ui_visibility(self, show: bool):
        self.ui.lGaussBT.setVisible(show)
        self.ui.lGaussWidth.setVisible(show)
        self.ui.spinBoxGaussBT.setVisible(show)
        self.ui.spinBoxGaussFilterWidth.setVisible(show)

        self.ui.spinBoxGaussFilterWidth.setValue(
            self.current_modulator.gauss_filter_width
        )
        self.ui.spinBoxGaussBT.setValue(self.current_modulator.gauss_bt)

    def closeEvent(self, event: QCloseEvent):
        self.ui.lineEditParameters.editingFinished.emit()
        settings.write(
            "{}/geometry".format(self.__class__.__name__), self.saveGeometry()
        )

        for gv in (
            self.ui.gVCarrier,
            self.ui.gVData,
            self.ui.gVModulated,
            self.ui.gVOriginalSignal,
        ):
            # Eliminate graphic views to prevent segfaults
            gv.eliminate()

        super().closeEvent(event)

    @property
    def current_modulator(self):
        return self.modulators[self.ui.comboBoxCustomModulations.currentIndex()]

    def set_ui_for_current_modulator(self):
        index = self.ui.comboBoxModulationType.findText(
            "*(" + self.current_modulator.modulation_type + ")",
            Qt.MatchFlag.MatchWildcard,
        )
        self.ui.comboBoxModulationType.setCurrentIndex(index)
        self.ui.doubleSpinBoxCarrierFreq.setValue(
            self.current_modulator.carrier_freq_hz
        )
        self.ui.doubleSpinBoxCarrierPhase.setValue(
            self.current_modulator.carrier_phase_deg
        )
        self.ui.spinBoxSamplesPerSymbol.setValue(
            self.current_modulator.samples_per_symbol
        )
        self.ui.spinBoxSampleRate.setValue(self.current_modulator.sample_rate)
        self.ui.spinBoxBitsPerSymbol.setValue(self.current_modulator.bits_per_symbol)

        self.update_modulation_parameters()

    def create_connects(self):
        self.ui.doubleSpinBoxCarrierFreq.valueChanged.connect(
            self.on_carrier_freq_changed
        )
        self.ui.doubleSpinBoxCarrierPhase.valueChanged.connect(
            self.on_carrier_phase_changed
        )
        self.ui.spinBoxSamplesPerSymbol.valueChanged.connect(
            self.on_samples_per_symbol_changed
        )
        self.ui.spinBoxSampleRate.valueChanged.connect(self.on_sample_rate_changed)
        self.ui.linEdDataBits.textChanged.connect(self.on_data_bits_changed)
        self.ui.spinBoxBitsPerSymbol.valueChanged.connect(
            self.on_bits_per_symbol_changed
        )
        self.ui.comboBoxModulationType.currentIndexChanged.connect(
            self.on_modulation_type_changed
        )
        self.ui.gVOriginalSignal.zoomed.connect(self.on_orig_signal_zoomed)
        self.ui.cbShowDataBitsOnly.stateChanged.connect(
            self.on_show_data_bits_only_changed
        )
        self.ui.btnSearchNext.clicked.connect(self.on_btn_next_search_result_clicked)
        self.ui.btnSearchPrev.clicked.connect(self.on_btn_prev_search_result_clicked)
        self.ui.comboBoxCustomModulations.editTextChanged.connect(
            self.on_custom_modulation_name_edited
        )
        self.ui.comboBoxCustomModulations.currentIndexChanged.connect(
            self.on_custom_modulation_index_changed
        )
        self.ui.btnAddModulation.clicked.connect(self.add_modulator)
        self.ui.btnRemoveModulation.clicked.connect(self.on_remove_modulator_clicked)
        self.ui.gVModulated.zoomed.connect(self.on_carrier_data_modulated_zoomed)
        self.ui.gVCarrier.zoomed.connect(self.on_carrier_data_modulated_zoomed)
        self.ui.gVData.zoomed.connect(self.on_carrier_data_modulated_zoomed)
        self.ui.gVModulated.selection_width_changed.connect(
            self.on_modulated_selection_changed
        )
        self.ui.gVOriginalSignal.selection_width_changed.connect(
            self.on_original_selection_changed
        )
        self.ui.spinBoxGaussBT.valueChanged.connect(self.on_gauss_bt_changed)
        self.ui.spinBoxGaussFilterWidth.valueChanged.connect(
            self.on_gauss_filter_width_changed
        )

        self.ui.chkBoxLockSIV.stateChanged.connect(self.on_lock_siv_changed)

        self.ui.gVOriginalSignal.signal_loaded.connect(self.handle_signal_loaded)
        self.ui.btnAutoDetect.clicked.connect(self.on_btn_autodetect_clicked)

        self.restore_bits_action.triggered.connect(
            self.on_restore_bits_action_triggered
        )
        self.configure_parameters_action.triggered.connect(
            self.on_configure_parameters_action_triggered
        )
        self.ui.lineEditParameters.editingFinished.connect(
            self.on_line_edit_parameters_editing_finished
        )

    def draw_carrier(self):
        self.ui.gVCarrier.plot_data(self.current_modulator.carrier_data)

    def draw_data_bits(self):
        self.ui.gVData.setScene(self.current_modulator.data_scene)
        self.ui.gVData.update()

    def draw_modulated(self):
        self.ui.gVModulated.plot_data(self.current_modulator.modulate(pause=0).imag)
        if self.lock_samples_in_view:
            siv = self.ui.gVOriginalSignal.view_rect().width()
            self.adjust_samples_in_view(siv)
        else:
            self.mark_samples_in_view()

    def draw_original_signal(self, start=0, end=-1):
        scene_manager = self.ui.gVOriginalSignal.scene_manager
        if scene_manager is None:
            return

        if end == -1:
            end = scene_manager.signal.num_samples

        y = self.ui.gVOriginalSignal.view_rect().y()
        h = self.ui.gVOriginalSignal.view_rect().height()
        self.ui.gVOriginalSignal.setSceneRect(start, y, end - start, h)
        self.ui.gVOriginalSignal.fitInView(self.ui.gVOriginalSignal.sceneRect())
        scene_manager.show_scene_section(start, end)
        self.ui.gVOriginalSignal.update()

        if self.lock_samples_in_view:
            self.adjust_samples_in_view(self.ui.gVModulated.view_rect().width())
        else:
            self.mark_samples_in_view()

    def update_views(self):
        self.ui.gVCarrier.update()
        self.ui.gVData.update()
        self.ui.gVModulated.update()
        self.ui.gVOriginalSignal.update()

    def search_data_sequence(self):
        if (
            not self.ui.cbShowDataBitsOnly.isEnabled()
            or not self.ui.cbShowDataBitsOnly.isChecked()
        ):
            return

        search_seq = self.ui.linEdDataBits.text()
        if len(search_seq) == 0 or self.protocol is None:
            return

        self.search_results[:] = []
        proto_bits = self.protocol.plain_bits_str
        len_seq = len(search_seq)

        for i, message in enumerate(proto_bits):
            j = message.find(search_seq)
            while j != -1:
                self.search_results.append((i, j, j + len_seq))
                j = message.find(search_seq, j + 1)

        self.ui.lTotalSearchresults.setText(str(len(self.search_results)))
        self.show_search_result(0)

    def show_search_result(self, i: int):
        if len(self.search_results) == 0:
            self.ui.lCurrentSearchResult.setText("0")
            self.ui.gVOriginalSignal.scene_manager.clear_path()
            return

        message, start_index, end_index = self.search_results[i]

        start, nsamples = self.protocol.get_samplepos_of_bitseq(
            message, start_index, message, end_index, False
        )
        self.draw_original_signal(start=start, end=start + nsamples)

        self.ui.lCurrentSearchResult.setText(str(i + 1))
        self.ui.btnSearchNext.setEnabled(i != len(self.search_results) - 1)
        self.ui.btnSearchPrev.setEnabled(i > 0)

    def add_modulator(self):
        names = [m.name for m in self.modulators]
        name = "Modulation"
        number = 1
        while name in names:
            name = "Modulation " + str(number)
            number += 1
        self.modulators.append(Modulator(name))
        self.ui.comboBoxCustomModulations.addItem(name)
        self.ui.comboBoxCustomModulations.setCurrentIndex(len(self.modulators) - 1)
        self.ui.btnRemoveModulation.setEnabled(True)

    def adjust_samples_in_view(self, target_siv: float):
        self.ui.gVOriginalSignal.scale(
            self.ui.gVOriginalSignal.view_rect().width() / target_siv, 1
        )
        mod_zoom_factor = self.ui.gVModulated.view_rect().width() / target_siv
        self.ui.gVModulated.scale(mod_zoom_factor, 1)
        self.ui.gVCarrier.scale(mod_zoom_factor, 1)
        self.ui.gVData.scale(mod_zoom_factor, 1)
        self.mark_samples_in_view()

    def detect_fsk_frequencies(self):
        if not self.current_modulator.is_frequency_based:
            return

        frequencies = []
        try:
            if not self.current_modulator.is_binary_modulation:
                raise NotImplementedError()

            zero_freq = self.protocol.estimate_frequency_for_zero(
                self.current_modulator.sample_rate
            )
            one_freq = self.protocol.estimate_frequency_for_one(
                self.current_modulator.sample_rate
            )
            zero_freq = self.__trim_number(zero_freq)
            one_freq = self.__trim_number(one_freq)
            zero_freq, one_freq = self.__ensure_multitude(zero_freq, one_freq)

            if zero_freq == one_freq:
                # If frequencies are equal, it is very likely the zero freq is negative
                zero_freq = -one_freq

            frequencies = [zero_freq, one_freq]

        except (AttributeError, NotImplementedError):
            frequencies = self.current_modulator.get_default_parameters()

        self.current_modulator.parameters = array("f", frequencies)
        self.update_modulation_parameters()

    def handle_signal_loaded(self, protocol):
        self.setCursor(Qt.CursorShape.WaitCursor)
        self.ui.cbShowDataBitsOnly.setEnabled(True)
        self.ui.chkBoxLockSIV.setEnabled(True)
        self.ui.btnAutoDetect.setEnabled(True)
        self.protocol = protocol

        # Apply bit length of original signal to current modulator
        self.ui.spinBoxSamplesPerSymbol.setValue(
            self.ui.gVOriginalSignal.signal.samples_per_symbol
        )

        # https://github.com/jopohl/urh/issues/130
        self.ui.gVModulated.show_full_scene(reinitialize=True)
        self.ui.gVCarrier.show_full_scene(reinitialize=True)
        self.ui.gVData.show_full_scene(reinitialize=True)

        self.unsetCursor()

    def mark_samples_in_view(self):
        self.ui.lSamplesInViewModulated.setText(
            str(int(self.ui.gVModulated.view_rect().width()))
        )

        if self.ui.gVOriginalSignal.scene_manager is not None:
            self.ui.lSamplesInViewOrigSignal.setText(
                str(int(self.ui.gVOriginalSignal.view_rect().width()))
            )
        else:
            self.ui.lSamplesInViewOrigSignal.setText("-")
            return

        if int(self.ui.gVOriginalSignal.view_rect().width()) != int(
            self.ui.gVModulated.view_rect().width()
        ):
            font = self.ui.lSamplesInViewModulated.font()
            font.setBold(False)
            self.ui.lSamplesInViewModulated.setFont(font)
            self.ui.lSamplesInViewOrigSignal.setFont(font)

            self.ui.lSamplesInViewOrigSignal.setStyleSheet("QLabel { color : red; }")
            self.ui.lSamplesInViewModulated.setStyleSheet("QLabel { color : red; }")
        else:
            font = self.ui.lSamplesInViewModulated.font()
            font.setBold(True)
            self.ui.lSamplesInViewModulated.setFont(font)
            self.ui.lSamplesInViewOrigSignal.setFont(font)

            self.ui.lSamplesInViewOrigSignal.setStyleSheet("")
            self.ui.lSamplesInViewModulated.setStyleSheet("")

    def set_default_modulation_parameters(self):
        self.current_modulator.parameters = (
            self.current_modulator.get_default_parameters()
        )
        self.update_modulation_parameters()

    def set_modulation_profile_status(self):
        visible = settings.read("multiple_modulations", False, bool)
        self.ui.btnAddModulation.setVisible(visible)
        self.ui.btnRemoveModulation.setVisible(visible)
        self.ui.comboBoxCustomModulations.setVisible(visible)

    def resizeEvent(self, event: QResizeEvent):
        self.update_views()

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key.Key_Enter or event.key() == Qt.Key.Key_Return:
            return
        else:
            super().keyPressEvent(event)

    def initialize(self, bits: str):
        self.on_modulation_type_changed()  # for drawing modulated signal initially
        self.original_bits = bits
        self.ui.linEdDataBits.setText(bits)
        self.draw_original_signal()
        self.ui.gVModulated.show_full_scene(reinitialize=True)
        self.ui.gVModulated.auto_fit_view()
        self.ui.gVData.show_full_scene(reinitialize=True)
        self.ui.gVData.auto_fit_view()
        self.ui.gVCarrier.show_full_scene(reinitialize=True)
        self.ui.gVCarrier.auto_fit_view()

        self.mark_samples_in_view()

    def update_modulation_parameters(self):
        n = len(self.current_modulator.parameters) - 1
        if self.current_modulator.is_amplitude_based:
            regex = r"(100|[0-9]{1,2})"
        elif self.current_modulator.is_frequency_based:
            regex = r"((-?[0-9]+)[.,]?[0-9]*[kKmMgG]?)"
        elif self.current_modulator.is_phase_based:
            regex = r"(-?(36[0]|3[0-5][0-9]|[12][0-9][0-9]|[1-9]?[0-9]))"
        else:
            raise ValueError("Unknown modulation type")

        full_regex = r"^(" + regex + r"/){" + str(n) + "}" + regex + r"$"
        self.ui.lineEditParameters.setValidator(
            QRegularExpressionValidator(QRegularExpression(full_regex))
        )
        self.ui.lineEditParameters.setText(self.current_modulator.parameters_string)

    def set_bits_per_symbol_enabled_status(self):
        if self.current_modulator.modulation_type == "OQPSK":
            self.ui.spinBoxBitsPerSymbol.setEnabled(False)
            self.ui.spinBoxBitsPerSymbol.setValue(2)
        else:
            self.ui.spinBoxBitsPerSymbol.setEnabled(True)

    def show_full_scene(self):
        for graphic_view in (self.ui.gVModulated, self.ui.gVData, self.ui.gVCarrier):
            graphic_view.show_full_scene(reinitialize=True)

    @pyqtSlot()
    def on_carrier_freq_changed(self):
        self.current_modulator.carrier_freq_hz = (
            self.ui.doubleSpinBoxCarrierFreq.value()
        )
        self.draw_carrier()
        self.draw_modulated()

    @pyqtSlot()
    def on_carrier_phase_changed(self):
        self.current_modulator.carrier_phase_deg = (
            self.ui.doubleSpinBoxCarrierPhase.value()
        )
        self.draw_carrier()
        self.draw_modulated()

    @pyqtSlot()
    def on_samples_per_symbol_changed(self):
        self.current_modulator.samples_per_symbol = (
            self.ui.spinBoxSamplesPerSymbol.value()
        )
        self.draw_carrier()
        self.draw_data_bits()
        self.draw_modulated()
        self.show_full_scene()

    @pyqtSlot()
    def on_data_bits_changed(self):
        text = self.ui.linEdDataBits.text()
        text = "".join(c for c in text if c == "1" or c == "0")
        self.ui.linEdDataBits.blockSignals(True)
        self.ui.linEdDataBits.setText(text)
        self.ui.linEdDataBits.blockSignals(False)
        self.current_modulator.display_bits = text
        self.draw_carrier()
        self.draw_data_bits()
        self.draw_modulated()
        if len(text) > 0:
            if len(text) > 24:
                display_text = text[0:24] + "..."
            else:
                display_text = text
            self.ui.cbShowDataBitsOnly.setToolTip(text)
            self.ui.cbShowDataBitsOnly.setText(
                self.tr("Show Only Data Sequence\n") + "(" + display_text + ")"
            )
        else:
            self.ui.cbShowDataBitsOnly.setToolTip("")
            self.ui.cbShowDataBitsOnly.setText(self.tr("Show Only Data Sequence\n"))

        self.search_data_sequence()
        self.restore_bits_action.setEnabled(text != self.original_bits)
        self.show_full_scene()

    @pyqtSlot()
    def on_sample_rate_changed(self):
        if int(self.ui.spinBoxSampleRate.value()) > 0:
            self.current_modulator.sample_rate = int(self.ui.spinBoxSampleRate.value())
            self.draw_carrier()
            self.draw_modulated()

    @pyqtSlot()
    def on_gauss_bt_changed(self):
        self.current_modulator.gauss_bt = self.ui.spinBoxGaussBT.value()
        self.draw_modulated()

    @pyqtSlot()
    def on_gauss_filter_width_changed(self):
        self.current_modulator.gauss_filter_width = (
            self.ui.spinBoxGaussFilterWidth.value()
        )
        self.draw_modulated()

    @pyqtSlot()
    def on_bits_per_symbol_changed(self):
        if (
            self.current_modulator.bits_per_symbol
            == self.ui.spinBoxBitsPerSymbol.value()
        ):
            return
        self.current_modulator.bits_per_symbol = self.ui.spinBoxBitsPerSymbol.value()
        self.set_default_modulation_parameters()
        self.draw_modulated()
        self.show_full_scene()

    @pyqtSlot()
    def on_modulation_type_changed(self):
        write_default_parameters = (
            self.current_modulator.modulation_type != self.__cur_selected_mod_type()
        )
        self.current_modulator.modulation_type = self.__cur_selected_mod_type()

        self.__set_gauss_ui_visibility(self.__cur_selected_mod_type() == "GFSK")

        self.ui.labelParameters.setText(self.current_modulator.parameter_type_str)
        if write_default_parameters:
            self.set_default_modulation_parameters()
        else:
            self.update_modulation_parameters()

        self.set_bits_per_symbol_enabled_status()
        self.draw_modulated()
        self.show_full_scene()

    @pyqtSlot()
    def on_orig_signal_zoomed(self):
        start = self.ui.gVOriginalSignal.view_rect().x()
        end = start + self.ui.gVOriginalSignal.view_rect().width()

        self.ui.gVOriginalSignal.centerOn(start + (end - start) / 2, 0)
        if self.lock_samples_in_view:
            self.adjust_samples_in_view(self.ui.gVOriginalSignal.view_rect().width())

            x = (
                self.ui.gVOriginalSignal.view_rect().x()
                + self.ui.gVOriginalSignal.view_rect().width() / 2
            )
            y = 0

            self.ui.gVModulated.centerOn(x, y)
            self.ui.gVCarrier.centerOn(x, y)
            self.ui.gVData.centerOn(x, y)
        else:
            self.mark_samples_in_view()

    @pyqtSlot(float)
    def on_carrier_data_modulated_zoomed(self, factor: float):
        x = self.sender().view_rect().x() + self.sender().view_rect().width() / 2
        y = 0
        for gv in (self.ui.gVCarrier, self.ui.gVData, self.ui.gVModulated):
            if gv == self.sender():
                continue
            if factor == -1:
                gv.show_full_scene()
            else:
                gv.scale(factor, 1)
                gv.centerOn(x, y)

        if self.lock_samples_in_view:
            self.adjust_samples_in_view(self.ui.gVModulated.view_rect().width())
            self.ui.gVOriginalSignal.centerOn(x, y)
        else:
            self.mark_samples_in_view()

    @pyqtSlot()
    def on_custom_modulation_name_edited(self):
        self.current_modulator.name = self.ui.comboBoxCustomModulations.currentText()

    @pyqtSlot()
    def on_custom_modulation_index_changed(self):
        self.set_ui_for_current_modulator()
        self.draw_carrier()
        self.draw_data_bits()
        self.draw_modulated()

    @pyqtSlot()
    def on_btn_next_search_result_clicked(self):
        cur_search_result = int(self.ui.lCurrentSearchResult.text()) - 1
        self.show_search_result(cur_search_result + 1)

    @pyqtSlot()
    def on_btn_prev_search_result_clicked(self):
        cur_search_result = int(self.ui.lCurrentSearchResult.text()) - 1
        self.show_search_result(cur_search_result - 1)

    @pyqtSlot()
    def on_show_data_bits_only_changed(self, redraw=True):
        show_data_bits_only = self.ui.cbShowDataBitsOnly.isChecked()
        if not self.ui.cbShowDataBitsOnly.isEnabled() or not show_data_bits_only:
            self.ui.btnSearchPrev.setEnabled(False)
            self.ui.btnSearchNext.setEnabled(False)
            self.ui.lCurrentSearchResult.setText("-")
            self.ui.lTotalSearchresults.setText("-")
        else:
            self.search_data_sequence()

        if not redraw:
            return

        if self.ui.cbShowDataBitsOnly.isEnabled() and not show_data_bits_only:
            self.draw_original_signal()

    @pyqtSlot()
    def on_remove_modulator_clicked(self):
        index = self.ui.comboBoxCustomModulations.currentIndex()
        self.ui.comboBoxCustomModulations.removeItem(index)
        self.modulators.remove(self.modulators[index])

        if len(self.modulators) == 1:
            self.ui.btnRemoveModulation.setDisabled(True)

    @pyqtSlot()
    def on_lock_siv_changed(self):
        self.lock_samples_in_view = self.ui.chkBoxLockSIV.isChecked()
        if self.lock_samples_in_view:
            self.adjust_samples_in_view(self.ui.gVModulated.view_rect().width())

    @pyqtSlot()
    def on_restore_bits_action_triggered(self):
        self.ui.linEdDataBits.setText(self.original_bits)

    @pyqtSlot()
    def on_btn_autodetect_clicked(self):
        signal = self.ui.gVOriginalSignal.scene_manager.signal
        freq = self.current_modulator.estimate_carrier_frequency(signal, self.protocol)

        if freq is None or freq == 0:
            QMessageBox.information(
                self,
                self.tr("No results"),
                self.tr("Unable to detect parameters from current signal"),
            )
            return

        self.ui.doubleSpinBoxCarrierFreq.setValue(freq)
        self.detect_fsk_frequencies()

    @pyqtSlot(int)
    def on_modulated_selection_changed(self, new_width: int):
        self.ui.lModulatedSelectedSamples.setText(str(abs(new_width)))

    @pyqtSlot(int)
    def on_original_selection_changed(self, new_width: int):
        self.ui.lOriginalSignalSamplesSelected.setText(str(abs(new_width)))

    @pyqtSlot()
    def on_configure_parameters_action_triggered(self):
        self.ui.lineEditParameters.editingFinished.emit()
        dialog = ModulationParametersDialog(
            self.current_modulator.parameters,
            self.current_modulator.modulation_type,
            self,
        )
        dialog.accepted.connect(self.update_modulation_parameters)
        dialog.show()

    @pyqtSlot()
    def on_line_edit_parameters_editing_finished(self):
        if not self.ui.lineEditParameters.hasAcceptableInput():
            return

        text = self.ui.lineEditParameters.text()
        parameters = []
        for param in text.split("/"):
            param = param.upper().replace(",", ".")
            factor = 1
            if param.endswith("G"):
                factor = 10**9
                param = param[:-1]
            elif param.endswith("M"):
                factor = 10**6
                param = param[:-1]
            elif param.endswith("K"):
                factor = 10**3
                param = param[:-1]

            try:
                parameters.append(factor * float(param))
            except ValueError:
                logger.warning("Could not convert {} to number".format(param))
                return

        self.current_modulator.parameters[:] = array("f", parameters)
        self.draw_modulated()
        self.show_full_scene()



import array
import copy
from collections import OrderedDict

from PyQt6.QtCore import (
    pyqtSlot,
    QAbstractTableModel,
    QModelIndex,
    Qt,
    QRegularExpression,
)
from PyQt6.QtGui import QRegularExpressionValidator
from PyQt6.QtWidgets import QWidget, QHeaderView, QAbstractItemView, QLineEdit

from urh.signalprocessing.ChecksumLabel import ChecksumLabel
from urh.signalprocessing.Message import Message
from urh.ui.delegates.SpinBoxDelegate import SpinBoxDelegate
from urh.ui.ui_checksum_options_widget import Ui_ChecksumOptions
from urh.util import util
from urh.util.GenericCRC import GenericCRC
from urh.util.Logger import logger
from urh.util.WSPChecksum import WSPChecksum


class ChecksumWidget(QWidget):
    SPECIAL_CRCS = OrderedDict(
        [
            ("CC1101", GenericCRC(polynomial="16_standard", start_value=True)),
        ]
    )

    class RangeTableModel(QAbstractTableModel):
        header_labels = ["Start", "End"]

        def __init__(
            self,
            checksum_label: ChecksumLabel,
            message: Message,
            proto_view: int,
            parent=None,
        ):
            """

            :param message:
            :type field_types: list of FieldType
            :param parent:
            """
            super().__init__(parent)
            self.checksum_label = checksum_label
            self.message = message
            self.proto_view = proto_view
            self.update()

        def update(self):
            self.beginResetModel()
            self.endResetModel()

        def columnCount(self, parent: QModelIndex = None, *args, **kwargs):
            return len(self.header_labels)

        def rowCount(self, parent: QModelIndex = None, *args, **kwargs):
            return len(self.checksum_label.data_ranges)

        def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
            if (
                role == Qt.ItemDataRole.DisplayRole
                and orientation == Qt.Orientation.Horizontal
            ):
                return self.header_labels[section]
            return super().headerData(section, orientation, role)

        def data(self, index: QModelIndex, role=Qt.ItemDataRole.DisplayRole):
            if not index.isValid():
                return None

            i, j = index.row(), index.column()

            if role == Qt.ItemDataRole.DisplayRole:
                data_range = self.checksum_label.data_ranges[i]
                if j == 0:
                    return (
                        self.message.convert_index(
                            data_range[0], 0, self.proto_view, True
                        )[0]
                        + 1
                    )
                elif j == 1:
                    return self.message.convert_index(
                        data_range[1], 0, self.proto_view, True
                    )[0]
            return None

        def setData(self, index: QModelIndex, value, role: int = ...):
            try:
                int_val = int(value)
            except ValueError:
                return False

            i, j = index.row(), index.column()

            if i > len(self.checksum_label.data_ranges):
                return False

            data_range = self.checksum_label.data_ranges[i]

            if j == 0:
                converted_index = self.message.convert_index(
                    int_val - 1, self.proto_view, 0, True
                )[0]
                if converted_index < data_range[1]:
                    data_range[0] = converted_index
            elif j == 1:
                converted_index = self.message.convert_index(
                    int_val, self.proto_view, 0, True
                )[0]
                if converted_index > data_range[0]:
                    data_range[1] = converted_index

            return True

        def flags(self, index):
            if not index.isValid():
                return Qt.ItemFlag.NoItemFlags

            try:
                _ = self.checksum_label.data_ranges[index.row()]
            except IndexError:
                return Qt.ItemFlag.NoItemFlags

            return (
                Qt.ItemFlag.ItemIsEditable
                | Qt.ItemFlag.ItemIsEnabled
                | Qt.ItemFlag.ItemIsSelectable
            )

    def __init__(
        self,
        checksum_label: ChecksumLabel,
        message: Message,
        proto_view: int,
        parent=None,
    ):
        super().__init__(parent)
        self.ui = Ui_ChecksumOptions()
        self.ui.setupUi(self)
        self.checksum_label = checksum_label
        self.data_range_table_model = self.RangeTableModel(
            checksum_label, message, proto_view, parent=self
        )
        self.ui.tableViewDataRanges.setItemDelegateForColumn(
            0, SpinBoxDelegate(1, 999999, self)
        )
        self.ui.tableViewDataRanges.setItemDelegateForColumn(
            1, SpinBoxDelegate(1, 999999, self)
        )
        self.ui.tableViewDataRanges.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeMode.Stretch
        )
        self.ui.tableViewDataRanges.setModel(self.data_range_table_model)
        self.ui.tableViewDataRanges.setEditTriggers(
            QAbstractItemView.EditTrigger.AllEditTriggers
        )
        self.display_crc_data_ranges_in_table()
        self.ui.comboBoxCRCFunction.addItems(
            [crc_name for crc_name in GenericCRC.DEFAULT_POLYNOMIALS]
        )
        self.ui.comboBoxCRCFunction.addItems(
            [special_crc_name for special_crc_name in self.SPECIAL_CRCS]
        )
        self.ui.lineEditCRCPolynomial.setValidator(
            QRegularExpressionValidator(QRegularExpression("[0-9,a-f]*"))
        )
        self.ui.comboBoxCategory.clear()
        for _, member in self.checksum_label.Category.__members__.items():
            self.ui.comboBoxCategory.addItem(member.value)
        self.set_ui_for_category()
        self.setFocus()
        self.create_connects()

    @property
    def proto_view(self):
        return self.data_range_table_model.proto_view

    @proto_view.setter
    def proto_view(self, value):
        if value != self.data_range_table_model.proto_view:
            self.data_range_table_model.proto_view = value
            self.data_range_table_model.update()

    def create_connects(self):
        self.ui.comboBoxCRCFunction.currentIndexChanged.connect(
            self.on_combobox_crc_function_current_index_changed
        )
        self.ui.btnAddRange.clicked.connect(self.on_btn_add_range_clicked)
        self.ui.btnRemoveRange.clicked.connect(self.on_btn_remove_range_clicked)
        self.ui.lineEditCRCPolynomial.editingFinished.connect(
            self.on_line_edit_crc_polynomial_editing_finished
        )
        self.ui.lineEditStartValue.editingFinished.connect(
            self.on_line_edit_start_value_editing_finished
        )
        self.ui.lineEditFinalXOR.editingFinished.connect(
            self.on_line_edit_final_xor_editing_finished
        )
        self.ui.comboBoxCategory.currentIndexChanged.connect(
            self.on_combobox_category_current_index_changed
        )
        self.ui.radioButtonWSPAuto.clicked.connect(
            self.on_radio_button_wsp_auto_clicked
        )
        self.ui.radioButtonWSPChecksum4.clicked.connect(
            self.on_radio_button_wsp_checksum4_clicked
        )
        self.ui.radioButtonWSPChecksum8.clicked.connect(
            self.on_radio_button_wsp_checksum8_clicked
        )
        self.ui.radioButtonWSPCRC8.clicked.connect(
            self.on_radio_button_wsp_crc8_clicked
        )
        self.ui.checkBoxRefIn.clicked.connect(self.on_check_box_ref_in_clicked)
        self.ui.checkBoxRefOut.clicked.connect(self.on_check_box_ref_out_clicked)

    def set_checksum_ui_elements(self):
        if self.checksum_label.is_generic_crc:
            self.ui.lineEditCRCPolynomial.setText(
                self.checksum_label.checksum.polynomial_as_hex_str
            )
            self.ui.lineEditStartValue.setText(
                util.bit2hex(self.checksum_label.checksum.start_value)
            )
            self.ui.lineEditFinalXOR.setText(
                util.bit2hex(self.checksum_label.checksum.final_xor)
            )
            self.ui.checkBoxRefIn.setChecked(self.checksum_label.checksum.lsb_first)
            self.ui.checkBoxRefOut.setChecked(self.checksum_label.checksum.reverse_all)
            self.__set_crc_function_index()
            self.__ensure_same_length()
            self.__set_crc_info_label()
        elif self.checksum_label.category == self.checksum_label.Category.wsp:
            if self.checksum_label.checksum.mode == WSPChecksum.ChecksumMode.auto:
                self.ui.radioButtonWSPAuto.setChecked(True)
            elif (
                self.checksum_label.checksum.mode == WSPChecksum.ChecksumMode.checksum4
            ):
                self.ui.radioButtonWSPChecksum4.setChecked(True)
            elif (
                self.checksum_label.checksum.mode == WSPChecksum.ChecksumMode.checksum8
            ):
                self.ui.radioButtonWSPChecksum8.setChecked(True)
            elif self.checksum_label.checksum.mode == WSPChecksum.ChecksumMode.crc8:
                self.ui.radioButtonWSPCRC8.setChecked(True)

    def set_ui_for_category(self):
        self.ui.comboBoxCategory.setCurrentText(self.checksum_label.category.value)
        if self.checksum_label.category == self.checksum_label.Category.generic:
            self.ui.stackedWidget.setCurrentWidget(self.ui.page_crc)
        elif self.checksum_label.category == self.checksum_label.Category.wsp:
            self.ui.stackedWidget.setCurrentWidget(self.ui.page_wsp)
        else:
            raise ValueError("Unknown category")

        self.set_checksum_ui_elements()

    def display_crc_data_ranges_in_table(self):
        self.data_range_table_model.update()

    def __set_crc_function_index(self):
        # Get the combobox index
        crc_found = False
        for crc_name in GenericCRC.DEFAULT_POLYNOMIALS:
            test_crc = GenericCRC(crc_name)
            if test_crc == self.checksum_label.checksum:
                self.ui.comboBoxCRCFunction.setCurrentText(crc_name)
                crc_found = True
                break

        if not crc_found:
            for crc_name, crc in self.SPECIAL_CRCS.items():
                if self.checksum_label.checksum == crc:
                    self.ui.comboBoxCRCFunction.setCurrentText(crc_name)
                    crc_found = True
                    break

        if not crc_found:
            self.__add_and_select_custom_item()
        elif "Custom" in [
            self.ui.comboBoxCRCFunction.itemText(i)
            for i in range(self.ui.comboBoxCRCFunction.count())
        ]:
            self.ui.comboBoxCRCFunction.removeItem(
                self.ui.comboBoxCRCFunction.count() - 1
            )

    def __set_crc_info_label(self):
        crc = self.checksum_label.checksum  # type: GenericCRC
        self.ui.label_crc_info.setText(
            "<b>CRC Summary:</b><ul>"
            "<li>Polynomial = {}<>"
            "<li>Length of checksum = {} bit</li>"
            "<li>start value length = {} bit</li>"
            "<li>final XOR length = {} bit</li>"
            "</ul>".format(
                crc.polynomial_to_html,
                crc.poly_order - 1,
                len(crc.start_value),
                len(crc.final_xor),
            )
        )

    def __ensure_same_length(self):
        for dependant_line_edit in [
            self.ui.lineEditStartValue,
            self.ui.lineEditFinalXOR,
        ]:  # type: QLineEdit
            if len(self.ui.lineEditCRCPolynomial.text()) < len(
                dependant_line_edit.text()
            ):
                dependant_line_edit.setText(
                    dependant_line_edit.text()[
                        : len(self.ui.lineEditCRCPolynomial.text())
                    ]
                )
                dependant_line_edit.editingFinished.emit()
            elif len(self.ui.lineEditCRCPolynomial.text()) > len(
                dependant_line_edit.text()
            ):
                # pad zeros at front
                dependant_line_edit.setText(
                    "0"
                    * (
                        len(self.ui.lineEditCRCPolynomial.text())
                        - len(dependant_line_edit.text())
                    )
                    + dependant_line_edit.text()
                )
                dependant_line_edit.editingFinished.emit()

    def __add_and_select_custom_item(self):
        if "Custom" not in [
            self.ui.comboBoxCRCFunction.itemText(i)
            for i in range(self.ui.comboBoxCRCFunction.count())
        ]:
            self.ui.comboBoxCRCFunction.addItem("Custom")
        self.ui.comboBoxCRCFunction.blockSignals(True)
        self.ui.comboBoxCRCFunction.setCurrentText("Custom")
        self.ui.comboBoxCRCFunction.blockSignals(False)

    @pyqtSlot()
    def on_btn_add_range_clicked(self):
        self.checksum_label.data_ranges.append([0, self.checksum_label.start])
        self.data_range_table_model.update()

    @pyqtSlot()
    def on_btn_remove_range_clicked(self):
        if len(self.checksum_label.data_ranges) > 1:
            self.checksum_label.data_ranges.pop(-1)
            self.data_range_table_model.update()

    @pyqtSlot(int)
    def on_combobox_crc_function_current_index_changed(self, index: int):
        poly_str = self.ui.comboBoxCRCFunction.itemText(index)
        if poly_str in GenericCRC.DEFAULT_POLYNOMIALS:
            self.checksum_label.checksum.polynomial = (
                self.checksum_label.checksum.choose_polynomial(poly_str)
            )
            self.checksum_label.checksum.start_value = array.array(
                "B", [0] * (self.checksum_label.checksum.poly_order - 1)
            )
            self.checksum_label.checksum.final_xor = array.array(
                "B", [0] * (self.checksum_label.checksum.poly_order - 1)
            )
        elif poly_str in self.SPECIAL_CRCS:
            self.checksum_label.checksum = copy.deepcopy(self.SPECIAL_CRCS[poly_str])
        else:
            logger.error("Unknown CRC")
            return

        self.ui.lineEditCRCPolynomial.setText(
            self.checksum_label.checksum.polynomial_as_hex_str
        )
        self.ui.lineEditStartValue.setText(
            util.bit2hex(self.checksum_label.checksum.start_value)
        )
        self.ui.lineEditFinalXOR.setText(
            util.bit2hex(self.checksum_label.checksum.final_xor)
        )
        self.ui.lineEditCRCPolynomial.editingFinished.emit()

    @pyqtSlot()
    def on_line_edit_crc_polynomial_editing_finished(self):
        self.checksum_label.checksum.set_polynomial_from_hex(
            self.ui.lineEditCRCPolynomial.text()
        )
        self.__ensure_same_length()
        self.__set_crc_info_label()
        self.__set_crc_function_index()

    @pyqtSlot()
    def on_check_box_ref_in_clicked(self):
        self.checksum_label.checksum.lsb_first = self.ui.checkBoxRefIn.isChecked()

    @pyqtSlot()
    def on_check_box_ref_out_clicked(self):
        self.checksum_label.checksum.reverse_all = self.ui.checkBoxRefOut.isChecked()

    @pyqtSlot()
    def on_line_edit_start_value_editing_finished(self):
        crc = self.checksum_label.checksum
        start_value = util.hex2bit(self.ui.lineEditStartValue.text())
        # pad with zeros at front
        start_value = (
            array.array("B", [0] * (crc.poly_order - 1 - len(start_value)))
            + start_value
        )
        crc.start_value = start_value[0 : crc.poly_order - 1]
        self.ui.lineEditStartValue.setText(util.bit2hex(crc.start_value))
        self.__set_crc_info_label()
        self.__set_crc_function_index()

    @pyqtSlot()
    def on_line_edit_final_xor_editing_finished(self):
        crc = self.checksum_label.checksum
        final_xor = util.hex2bit(self.ui.lineEditFinalXOR.text())
        final_xor = (
            array.array("B", [0] * (crc.poly_order - 1 - len(final_xor))) + final_xor
        )
        crc.final_xor = final_xor[0 : crc.poly_order - 1]
        self.ui.lineEditFinalXOR.setText(util.bit2hex(crc.final_xor))
        self.__set_crc_info_label()
        self.__set_crc_function_index()

    @pyqtSlot(int)
    def on_combobox_category_current_index_changed(self, index: int):
        self.checksum_label.category = self.checksum_label.Category(
            self.ui.comboBoxCategory.currentText()
        )
        self.set_ui_for_category()

    @pyqtSlot()
    def on_radio_button_wsp_auto_clicked(self):
        self.checksum_label.checksum.mode = WSPChecksum.ChecksumMode.auto

    @pyqtSlot()
    def on_radio_button_wsp_checksum4_clicked(self):
        self.checksum_label.checksum.mode = WSPChecksum.ChecksumMode.checksum4

    @pyqtSlot()
    def on_radio_button_wsp_checksum8_clicked(self):
        self.checksum_label.checksum.mode = WSPChecksum.ChecksumMode.checksum8

    @pyqtSlot()
    def on_radio_button_wsp_crc8_clicked(self):
        self.checksum_label.checksum.mode = WSPChecksum.ChecksumMode.crc8



from PyQt6.QtCore import QAbstractListModel, Qt, QModelIndex
from PyQt6.QtGui import QColor

from urh import settings


class ParticipantLegendListModel(QAbstractListModel):
    def __init__(self, participants, parent=None):
        """

        :type participants: list of Participant
        """
        super().__init__(parent)
        self.participants = participants

    def rowCount(self, QModelIndex_parent=None, *args, **kwargs):
        return len(self.participants) + 1

    def data(self, index: QModelIndex, role=None):
        row = index.row()
        if role == Qt.ItemDataRole.DisplayRole:
            if row == 0:
                return "not assigned"
            else:
                try:
                    return str(self.participants[row - 1])
                except IndexError:
                    return None
        elif role == Qt.ItemDataRole.BackgroundRole:
            if row > 0:
                try:
                    return settings.PARTICIPANT_COLORS[
                        self.participants[row - 1].color_index
                    ]
                except IndexError:
                    return None
        elif role == Qt.ItemDataRole.ForegroundRole:
            if row > 0:
                try:
                    bgcolor = settings.PARTICIPANT_COLORS[
                        self.participants[row - 1].color_index
                    ]
                    red, green, blue = bgcolor.red(), bgcolor.green(), bgcolor.blue()
                    return (
                        QColor("black")
                        if (red * 0.299 + green * 0.587 + blue * 0.114) > 186
                        else QColor("white")
                    )
                except IndexError:
                    return None

    def flags(self, index):
        return Qt.ItemFlag.ItemIsEnabled

    def update(self):
        self.beginResetModel()
        self.endResetModel()



import os
import shutil
import tarfile
import tempfile
import zipfile

import numpy as np
from PyQt6.QtCore import QDir
from PyQt6.QtWidgets import QFileDialog

from urh.signalprocessing.IQArray import IQArray
from urh.util.Errors import Errors

archives = {}
""":type: dict of [str, str]
   :param: archives[extracted_filename] = filename"""

RECENT_PATH = QDir.homePath()

SIGNAL_FILE_EXTENSIONS_BY_TYPE = {
    np.int8: ".complex16s",
    np.uint8: ".complex16u",
    np.int16: ".complex32s",
    np.uint16: ".complex32u",
    np.float32: ".complex",
    np.complex64: ".complex",
}

SIGNAL_NAME_FILTERS_BY_TYPE = {
    np.int8: "Complex16 signed (*.complex16s *.cs8)",
    np.uint8: "Complex16 unsigned (*.complex16u *.cu8)",
    np.uint16: "Complex32 unsigned (*.complex32u *.cu16)",
    np.int16: "Complex32 signed (*.complex32s *.cs16)",
    np.float32: "Complex (*.complex)",
    np.complex64: "Complex (*.complex)",
}

EVERYTHING_FILE_FILTER = "All Files (*)"

SIGNAL_NAME_FILTERS = list(sorted(set(SIGNAL_NAME_FILTERS_BY_TYPE.values())))

COMPRESSED_COMPLEX_FILE_FILTER = "Compressed Complex File (*.coco)"
WAV_FILE_FILTER = "Waveform Audio File Format (*.wav *.wave)"
PROTOCOL_FILE_FILTER = "Protocol (*.proto.xml *.proto)"
BINARY_PROTOCOL_FILE_FILTER = "Binary Protocol (*.bin)"
WIRESHARK_FILE_FILTER = "Wireshark File (*.pcapng)"
PLAIN_BITS_FILE_FILTER = "Plain Bits (*.txt)"
FUZZING_FILE_FILTER = "Fuzzing Profile (*.fuzz.xml *.fuzz)"
SIMULATOR_FILE_FILTER = "Simulator Profile (*.sim.xml *.sim)"
TAR_FILE_FILTER = "Tar Archive (*.tar *.tar.gz *.tar.bz2)"
ZIP_FILE_FILTER = "Zip Archive (*.zip)"
SUB_FILE_FILTER = "Flipper Zero SubGHz RAW (*.sub)"


def __get__name_filter_for_signals() -> str:
    return ";;".join(
        [EVERYTHING_FILE_FILTER]
        + SIGNAL_NAME_FILTERS
        + [COMPRESSED_COMPLEX_FILE_FILTER, WAV_FILE_FILTER, SUB_FILE_FILTER]
    )


def get_open_dialog(
    directory_mode=False, parent=None, name_filter="full"
) -> QFileDialog:
    dialog = QFileDialog(parent=parent, directory=RECENT_PATH)

    if directory_mode:
        dialog.setFileMode(QFileDialog.FileMode.Directory)
        dialog.setWindowTitle("Open Folder")
    else:
        dialog.setFileMode(QFileDialog.FileMode.ExistingFiles)
        dialog.setWindowTitle("Open Files")
        if name_filter == "full":
            name_filter = (
                __get__name_filter_for_signals()
                + ";;"
                + ";;".join(
                    [
                        PROTOCOL_FILE_FILTER,
                        BINARY_PROTOCOL_FILE_FILTER,
                        PLAIN_BITS_FILE_FILTER,
                        FUZZING_FILE_FILTER,
                        SIMULATOR_FILE_FILTER,
                        TAR_FILE_FILTER,
                        ZIP_FILE_FILTER,
                    ]
                )
            )
        elif name_filter == "signals_only":
            name_filter = __get__name_filter_for_signals()
        elif name_filter == "proto":
            name_filter = ";;".join([PROTOCOL_FILE_FILTER, BINARY_PROTOCOL_FILE_FILTER])
        elif name_filter == "fuzz":
            name_filter = FUZZING_FILE_FILTER
        elif name_filter == "simulator":
            name_filter = SIMULATOR_FILE_FILTER

        dialog.setNameFilter(name_filter)

    return dialog


def ask_save_file_name(
    initial_name: str, caption="Save signal", selected_name_filter=None
):
    global RECENT_PATH
    if caption == "Save signal":
        name_filter = __get__name_filter_for_signals()
    elif caption == "Save fuzzing profile":
        name_filter = FUZZING_FILE_FILTER
    elif caption == "Save encoding":
        name_filter = ""
    elif caption == "Save simulator profile":
        name_filter = SIMULATOR_FILE_FILTER
    elif caption == "Export spectrogram":
        name_filter = "Frequency Time (*.ft);;Frequency Time Amplitude (*.fta)"
    elif caption == "Save protocol":
        name_filter = ";;".join(
            [PROTOCOL_FILE_FILTER, BINARY_PROTOCOL_FILE_FILTER, WIRESHARK_FILE_FILTER]
        )
    elif caption == "Export demodulated":
        name_filter = ";;".join([WAV_FILE_FILTER, SUB_FILE_FILTER])
    else:
        name_filter = EVERYTHING_FILE_FILTER

    filename = None
    dialog = QFileDialog(directory=RECENT_PATH, caption=caption, filter=name_filter)
    dialog.setFileMode(QFileDialog.FileMode.AnyFile)
    dialog.setLabelText(QFileDialog.DialogLabel.Accept, "Save")
    dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)

    if selected_name_filter is not None:
        dialog.selectNameFilter(selected_name_filter)

    dialog.selectFile(initial_name)

    if dialog.exec():
        filename = dialog.selectedFiles()[0]

    if filename:
        RECENT_PATH = os.path.split(filename)[0]

    return filename


def ask_signal_file_name_and_save(
    signal_name: str, data, sample_rate=1e6, wav_only=False, parent=None
) -> str:
    if wav_only:
        if not signal_name.endswith(".wav") and not signal_name.endswith(".wave"):
            signal_name += ".wav"
        selected_name_filter = WAV_FILE_FILTER
    else:
        if not any(
            signal_name.endswith(e) for e in SIGNAL_NAME_FILTERS_BY_TYPE.values()
        ):
            try:
                dtype = next(
                    d for d in SIGNAL_FILE_EXTENSIONS_BY_TYPE.keys() if d == data.dtype
                )
                signal_name += SIGNAL_FILE_EXTENSIONS_BY_TYPE[dtype]
                selected_name_filter = SIGNAL_NAME_FILTERS_BY_TYPE[dtype]
            except StopIteration:
                selected_name_filter = None
        else:
            selected_name_filter = None

    filename = ask_save_file_name(
        signal_name, selected_name_filter=selected_name_filter
    )

    if filename:
        try:
            save_data(data, filename, sample_rate=sample_rate)
        except Exception as e:
            Errors.exception(e)
            filename = None
    else:
        filename = None

    return filename


def save_data(data, filename: str, sample_rate=1e6, num_channels=2):
    if not isinstance(data, IQArray):
        data = IQArray(data)

    if filename.endswith(".wav"):
        data.export_to_wav(filename, num_channels, sample_rate)
    elif filename.endswith(".coco"):
        data.save_compressed(filename)
    elif filename.endswith(".sub"):
        data.export_to_sub(filename, 433920000, "FuriHalSubGhzPresetOok650Async")
    else:
        data.tofile(filename)

    if filename in archives.keys():
        archive = archives[filename]
        if archive.endswith("zip"):
            rewrite_zip(archive)
        elif (
            archive.endswith("tar") or archive.endswith("bz2") or archive.endswith("gz")
        ):
            rewrite_tar(archive)


def save_signal(signal):
    save_data(signal.iq_array.data, signal.filename, signal.sample_rate)


def rewrite_zip(zip_name):
    tempdir = tempfile.mkdtemp()
    try:
        temp_name = os.path.join(tempdir, "new.zip")
        files_in_archive = [f for f in archives.keys() if archives[f] == zip_name]
        with zipfile.ZipFile(temp_name, "w") as zip_write:
            for filename in files_in_archive:
                zip_write.write(filename)
        shutil.move(temp_name, zip_name)
    finally:
        shutil.rmtree(tempdir)


def rewrite_tar(tar_name: str):
    tempdir = tempfile.mkdtemp()
    compression = ""
    if tar_name.endswith("gz"):
        compression = "gz"
    elif tar_name.endswith("bz2"):
        compression = "bz2"
    try:
        ext = "" if len(compression) == 0 else "." + compression
        temp_name = os.path.join(tempdir, "new.tar" + ext)
        files_in_archive = [f for f in archives.keys() if archives[f] == tar_name]
        with tarfile.open(temp_name, "w:" + compression) as tar_write:
            for file in files_in_archive:
                tar_write.add(file)
        shutil.move(temp_name, tar_name)
    finally:
        shutil.rmtree(tempdir)


def uncompress_archives(file_names, temp_dir):
    """
    Extract each archive from the list of filenames.
    Normal files stay untouched.
    Add all files to the Recent Files.
    :type file_names: list of str
    :type temp_dir: str
    :rtype: list of str
    """
    result = []
    for filename in file_names:
        if (
            filename.endswith(".tar")
            or filename.endswith(".tar.gz")
            or filename.endswith(".tar.bz2")
        ):
            obj = tarfile.open(filename, "r")
            extracted_file_names = []
            for j, member in enumerate(obj.getmembers()):
                obj.extract(member, temp_dir)
                extracted_filename = os.path.join(temp_dir, obj.getnames()[j])
                extracted_file_names.append(extracted_filename)
                archives[extracted_filename] = filename
            result.extend(extracted_file_names[:])
        elif filename.endswith(".zip"):
            obj = zipfile.ZipFile(filename)
            extracted_file_names = []
            for j, info in enumerate(obj.infolist()):
                obj.extract(info, path=temp_dir)
                extracted_filename = os.path.join(temp_dir, obj.namelist()[j])
                extracted_file_names.append(extracted_filename)
                archives[extracted_filename] = filename
            result.extend(extracted_file_names[:])
        else:
            result.append(filename)

    return result


def get_directory():
    directory = QFileDialog.getExistingDirectory(
        None,
        "Choose Directory",
        QDir.homePath(),
        QFileDialog.Option.ShowDirsOnly | QFileDialog.Option.DontResolveSymlinks,
    )
    return directory



from PyQt6.QtCore import QPoint, pyqtSignal, pyqtSlot
from PyQt6.QtWidgets import QWidget, QSizePolicy, QCheckBox, QMessageBox
from PyQt6.QtGui import QUndoStack

from urh import settings
from urh.controller.widgets.SignalFrame import SignalFrame
from urh.signalprocessing.Signal import Signal
from urh.ui.ui_tab_interpretation import Ui_Interpretation
from urh.util import util


class SignalTabController(QWidget):
    frame_closed = pyqtSignal(SignalFrame)
    not_show_again_changed = pyqtSignal()
    signal_created = pyqtSignal(int, Signal)
    files_dropped = pyqtSignal(list)
    frame_was_dropped = pyqtSignal(int, int)

    @property
    def num_frames(self):
        return len(self.signal_frames)

    @property
    def signal_frames(self):
        """

        :rtype: list of SignalFrame
        """
        splitter = self.ui.splitter
        return [
            splitter.widget(i)
            for i in range(splitter.count())
            if isinstance(splitter.widget(i), SignalFrame)
        ]

    @property
    def signal_undo_stack(self):
        return self.undo_stack

    def __init__(self, project_manager, parent=None):
        super().__init__(parent)
        self.ui = Ui_Interpretation()
        self.ui.setupUi(self)

        util.set_splitter_stylesheet(self.ui.splitter)

        self.ui.placeholderLabel.setVisible(False)
        self.getting_started_status = None
        self.__set_getting_started_status(True)

        self.undo_stack = QUndoStack()
        self.project_manager = project_manager

        self.drag_pos = None

    def on_files_dropped(self, files):
        self.files_dropped.emit(files)

    def close_frame(self, frame: SignalFrame):
        self.frame_closed.emit(frame)

    def add_signal_frame(self, proto_analyzer, index=-1):
        self.__set_getting_started_status(False)
        sig_frame = SignalFrame(
            proto_analyzer, self.undo_stack, self.project_manager, parent=self
        )
        sframes = self.signal_frames

        if len(proto_analyzer.signal.filename) == 0:
            # new signal from "create signal from selection"
            sig_frame.ui.btnSaveSignal.show()

        self.__create_connects_for_signal_frame(signal_frame=sig_frame)
        sig_frame.signal_created.connect(self.emit_signal_created)
        sig_frame.not_show_again_changed.connect(self.not_show_again_changed.emit)
        sig_frame.ui.lineEditSignalName.setToolTip(
            self.tr("Sourcefile: ") + proto_analyzer.signal.filename
        )
        sig_frame.apply_to_all_clicked.connect(self.on_apply_to_all_clicked)

        prev_signal_frame = sframes[-1] if len(sframes) > 0 else None
        if prev_signal_frame is not None and hasattr(prev_signal_frame, "ui"):
            sig_frame.ui.cbProtoView.setCurrentIndex(
                prev_signal_frame.ui.cbProtoView.currentIndex()
            )

        sig_frame.blockSignals(True)

        index = self.num_frames if index == -1 else index
        sig_frame.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding
        )
        self.ui.splitter.insertWidget(index, sig_frame)
        sig_frame.blockSignals(False)

        default_view = settings.read("default_view", 0, int)
        sig_frame.ui.cbProtoView.setCurrentIndex(default_view)

        return sig_frame

    def add_empty_frame(self, filename: str, proto):
        self.__set_getting_started_status(False)
        sig_frame = SignalFrame(
            proto_analyzer=proto,
            undo_stack=self.undo_stack,
            project_manager=self.project_manager,
            parent=self,
        )

        sig_frame.ui.lineEditSignalName.setText(filename)
        self.__create_connects_for_signal_frame(signal_frame=sig_frame)

        self.ui.splitter.insertWidget(self.num_frames, sig_frame)

        return sig_frame

    def __set_getting_started_status(self, getting_started: bool):
        if getting_started == self.getting_started_status:
            return

        self.getting_started_status = getting_started
        self.ui.labelGettingStarted.setVisible(getting_started)

        if not getting_started:
            w = QWidget()
            w.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding)
            self.ui.splitter.addWidget(w)

    def __create_connects_for_signal_frame(self, signal_frame: SignalFrame):
        signal_frame.hold_shift = settings.read("hold_shift_to_drag", True, type=bool)
        signal_frame.drag_started.connect(self.frame_dragged)
        signal_frame.frame_dropped.connect(self.frame_dropped)
        signal_frame.files_dropped.connect(self.on_files_dropped)
        signal_frame.closed.connect(self.close_frame)

    def set_frame_numbers(self):
        for i, f in enumerate(self.signal_frames):
            f.ui.lSignalNr.setText("{0:d}:".format(i + 1))

    @pyqtSlot()
    def save_all(self):
        if self.num_frames == 0:
            return

        try:
            not_show = settings.read("not_show_save_dialog", False, type=bool)
        except TypeError:
            not_show = False

        if not not_show:
            cb = QCheckBox("Don't ask me again.")
            msg_box = QMessageBox(
                QMessageBox.Icon.Question,
                self.tr("Confirm saving all signals"),
                self.tr("All changed signal files will be overwritten. OK?"),
            )
            msg_box.addButton(QMessageBox.StandardButton.Yes)
            msg_box.addButton(QMessageBox.StandardButton.No)
            msg_box.setCheckBox(cb)

            reply = msg_box.exec()
            not_show_again = cb.isChecked()
            settings.write("not_show_save_dialog", not_show_again)
            self.not_show_again_changed.emit()

            if reply != QMessageBox.StandardButton.Yes:
                return

        for f in self.signal_frames:
            if f.signal is None or f.signal.filename == "":
                continue
            f.signal.save()

    @pyqtSlot()
    def close_all(self):
        for f in self.signal_frames:
            f.my_close()

    @pyqtSlot(Signal)
    def on_apply_to_all_clicked(self, signal: Signal):
        for frame in self.signal_frames:
            if frame.signal is not None:
                frame.signal.noise_min_plot = signal.noise_min_plot
                frame.signal.noise_max_plot = signal.noise_max_plot

                frame.signal.block_protocol_update = True
                proto_needs_update = False

                if frame.signal.modulation_type != signal.modulation_type:
                    frame.signal.modulation_type = signal.modulation_type
                    proto_needs_update = True

                if frame.signal.center != signal.center:
                    frame.signal.center = signal.center
                    proto_needs_update = True

                if frame.signal.tolerance != signal.tolerance:
                    frame.signal.tolerance = signal.tolerance
                    proto_needs_update = True

                if frame.signal.noise_threshold != signal.noise_threshold:
                    frame.signal.noise_threshold_relative = (
                        signal.noise_threshold_relative
                    )
                    proto_needs_update = True

                if frame.signal.samples_per_symbol != signal.samples_per_symbol:
                    frame.signal.samples_per_symbol = signal.samples_per_symbol
                    proto_needs_update = True

                if frame.signal.pause_threshold != signal.pause_threshold:
                    frame.signal.pause_threshold = signal.pause_threshold
                    proto_needs_update = True

                if frame.signal.message_length_divisor != signal.message_length_divisor:
                    frame.signal.message_length_divisor = signal.message_length_divisor
                    proto_needs_update = True

                frame.signal.block_protocol_update = False

                if proto_needs_update:
                    frame.signal.protocol_needs_update.emit()

    @pyqtSlot(QPoint)
    def frame_dragged(self, pos: QPoint):
        self.drag_pos = pos

    @pyqtSlot(QPoint)
    def frame_dropped(self, pos: QPoint):
        start = self.drag_pos
        if start is None:
            return

        end = pos
        start_index = -1
        end_index = -1
        if self.num_frames > 1:
            for i, w in enumerate(self.signal_frames):
                if w.geometry().contains(start):
                    start_index = i

                if w.geometry().contains(end):
                    end_index = i

        self.swap_frames(start_index, end_index)
        self.frame_was_dropped.emit(start_index, end_index)

    @pyqtSlot(int, int)
    def swap_frames(self, from_index: int, to_index: int):
        if from_index != to_index:
            start_sig_widget = self.ui.splitter.widget(from_index)
            self.ui.splitter.insertWidget(to_index, start_sig_widget)

    @pyqtSlot()
    def on_participant_changed(self):
        for sframe in self.signal_frames:
            sframe.on_participant_changed()

    def redraw_spectrograms(self):
        for frame in self.signal_frames:
            if frame.ui.gvSpectrogram.width_spectrogram > 0:
                frame.draw_spectrogram(force_redraw=True)

    @pyqtSlot(Signal)
    def emit_signal_created(self, signal):
        try:
            index = self.signal_frames.index(self.sender()) + 1
        except ValueError:
            index = -1

        self.signal_created.emit(index, signal)



from PyQt6.QtGui import QKeySequence, QAction
from PyQt6.QtWidgets import QGraphicsView, QMenu

from urh.ui.views.SimulatorGraphicsView import SimulatorGraphicsView


class LoggingGraphicsView(QGraphicsView):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setDragMode(QGraphicsView.DragMode.RubberBandDrag)

        self.log_selected_action = QAction(self.tr("Log selected items"), self)
        self.log_selected_action.setShortcut(QKeySequence("L"))
        self.log_selected_action.triggered.connect(
            self.on_log_selected_action_triggered
        )

        self.do_not_log_selected_action = QAction(
            self.tr("Do not log selected items"), self
        )
        self.do_not_log_selected_action.setShortcut(QKeySequence("N"))
        self.do_not_log_selected_action.triggered.connect(
            self.on_do_not_log_selected_action_triggered
        )

        self.select_all_action = QAction(self.tr("Select all"), self)
        self.select_all_action.setShortcut(QKeySequence.StandardKey.SelectAll)
        self.select_all_action.triggered.connect(self.on_select_all_action_triggered)

        self.addAction(self.log_selected_action)
        self.addAction(self.do_not_log_selected_action)
        self.addAction(self.select_all_action)

    def contextMenuEvent(self, event):
        menu = self.create_context_menu()
        menu.exec(event.globalPos())

    def create_context_menu(self):
        menu = QMenu()

        if len(self.scene().selectedItems()):
            menu.addAction(self.log_selected_action)
            menu.addAction(self.do_not_log_selected_action)

        SimulatorGraphicsView.add_select_actions_to_menu(
            menu,
            self.scene(),
            select_to_trigger=self.on_select_to_action_triggered,
            select_from_trigger=self.on_select_from_action_triggered,
        )

        return menu

    def on_select_from_action_triggered(self):
        self.scene().select_messages_with_participant(self.sender().data())

    def on_select_to_action_triggered(self):
        self.scene().select_messages_with_participant(
            self.sender().data(), from_part=False
        )

    def on_log_selected_action_triggered(self):
        self.scene().log_selected_items(True)

    def on_do_not_log_selected_action_triggered(self):
        self.scene().log_selected_items(False)

    def on_select_all_action_triggered(self):
        self.scene().select_all_items()



import socket
import time

import numpy
import numpy as np

from urh.dev.gr.AbstractBaseThread import AbstractBaseThread
from urh.util import util
from urh.util.Logger import logger


class SenderThread(AbstractBaseThread):
    MAX_SAMPLES_PER_TRANSMISSION = 4096

    def __init__(
        self,
        frequency,
        sample_rate,
        bandwidth,
        gain,
        if_gain,
        baseband_gain,
        ip="127.0.0.1",
        parent=None,
    ):
        super().__init__(
            frequency,
            sample_rate,
            bandwidth,
            gain,
            if_gain,
            baseband_gain,
            False,
            ip,
            parent,
        )

        self.data = numpy.empty(1, dtype=numpy.complex64)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        port = util.get_free_port()
        self.gr_port = port

        self.sending_repeats = 1  # How often shall we send the data?

        self.__samples_per_transmission = self.MAX_SAMPLES_PER_TRANSMISSION

    @property
    def repeat_endless(self):
        return self.sending_repeats == 0 or self.sending_repeats == -1

    @property
    def samples_per_transmission(self):
        return self.__samples_per_transmission

    @samples_per_transmission.setter
    def samples_per_transmission(self, val: int):
        if val >= self.MAX_SAMPLES_PER_TRANSMISSION:
            self.__samples_per_transmission = self.MAX_SAMPLES_PER_TRANSMISSION
        elif val <= 1:
            self.__samples_per_transmission = 1
        else:
            self.__samples_per_transmission = 2 ** (int(np.log2(val)) - 1)

    def run(self):
        self.initialize_process()
        len_data = len(self.data)
        self.current_iteration = (
            self.current_iteration if self.current_iteration is not None else 0
        )
        time.sleep(1)

        try:
            while self.current_index < len_data and not self.isInterruptionRequested():
                time.sleep(self.samples_per_transmission / self.sample_rate)
                self.socket.sendto(
                    self.data[
                        self.current_index : self.current_index
                        + self.samples_per_transmission
                    ].tostring(),
                    (self.ip, self.gr_port),
                )
                self.current_index += self.samples_per_transmission

                if self.current_index >= len_data:
                    self.current_iteration += 1
                else:
                    continue

                if self.repeat_endless or self.current_iteration < self.sending_repeats:
                    self.current_index = 0

            self.current_index = len_data - 1
            self.current_iteration = None
            self.stop("FIN - All data was sent successfully")
        except RuntimeError:
            logger.error("Sender thread crashed.")
